//Ethan Alexander Shulman 2022 "Plastic Rain" February 7, 2022

//scene rendering options
#define VIEW_STEPS 100
#define VIEW_STEP_DITHER .1
#define VIEW_EPSILON 1e-4
#define VIEW_EPSILON_MUL 1.1
#define LIGHTING_STEPS 40
#define LIGHTING_STEP_DITHER .1
#define LIGHTING_EPSILON 1e-2
//#define COLOR_MAX 100.
#define DENOISE_RANGE 4
#define DENOISE_NORMAL_THRESHOLD .99
#define DENOISE_DEPTH_THRESHOLD 1e3
#define DENOISE_HASH_DIVIDE 1
#define ANIMATION_TIME 1.
#define ANIMATION_SAMPLES 50
#ifdef ANIMATION_TIME
out float time;
#endif

//constants
#define PI 3.14159265





mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))



/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	return mat3(0,5,-10,
			0,0,0,
			1.,0.,0.);
}


//distance dithering for volumetrics
#define START_VOLUME(posSeed) vec2 opacityRand = fract(baseRand.zw+128.*dot(abs(posSeed),baseRand.zxy)*baseRand.xy)
#define VOLUME(dst,skip,alpha,anisotropy) max(dst,-1e3+(skip+1e3)*step(alpha+nsampling*anisotropy,opacityRand.x))

//dither materials for chromatic separation
#define START_CHROMA(posSeed) float chromaFrame = mod(baseRand.x+156.*dot(abs(posSeed),baseRand.zwy),3.)
#define CHROMA(mt,rc,gc,bc) if (chromaFrame < 1.) {mt[0]*=vec3(3,0,0);rc;}\
else if (chromaFrame < 2.) {mt[0]*=vec3(0,3,0);gc;}\
else {mt[0]*=vec3(0,0,3);bc;}



//functions defining scene

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//sand
	vec2 geo = vec2(p.y,0);
	if (geo.x < 2.) {
		for (float f = 1.; f < 16.; f += f) {
			geo.x += cos(dot(p.xz,vec2(cos(f),sin(f))*f*.1))/f;
		}
		geo.x = max(geo.x,length(mod(abs(p+cos(p.zxy*30.)*.01),.01)-.005)-.005);
	}
	
	//structure
	float sd = 0.;
	p.y -= 4.;
	for (float s = 1.; s < 8.; s++){
		float sz = s, sz2 = sz*.5;
		vec3 rp = mod(abs(p+vec3(0,time*s,0)),sz)-sz2;
		sz2 *= .5;
		sd = max(sd,min(length(max(abs(rp.xy)-sz2,0.)),min(length(max(abs(rp.xz)-sz2,0.)),length(max(abs(rp.zy)-sz2,0.)))));
		p.xz *= r2d(.5);
	}
	if (sd < geo.x) geo = vec2(sd,1);
	
	return geo;
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;
float refractionIndex;*/
mat3 material(vec3 p, vec3 n, float mat) {
	//structure
	if (mat == 1.) {
		return mat3(vec3(.32,.33,.33), .1,0,1, 1,0,0);
	}
	
	//sand
	return mat3(vec3(.88,.75,.35), .2,0,1, 1,0,0);
}

//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return vec3(.23,.44,.8)*max(0.,d.y)+max(0.,1.-10.*length(d-normalize(vec3(1))))*500.;
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return background(p,d);
}

