#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2022

in vec3 in_Position;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};

vec4 baseRand;
float nsampling;

#include "art/PathTracerScene.glsl"


vec3 normal(vec3 p) {
	nsampling = 1.;
	vec3 n = normalize(vec3(geometry(p+vec3(VIEW_EPSILON,0,0)).x,geometry(p+vec3(0,VIEW_EPSILON,0)).x,geometry(p+vec3(0,0,VIEW_EPSILON)).x)-geometry(p).x);
	nsampling = 0.;
	if (n != n) return vec3(0,1,0);
	return n;
}


void vert() {
	#ifdef ANIMATION_TIME
	time = mod(float(frame)/float(framerate*ANIMATION_SAMPLES),ANIMATION_TIME);
	#endif
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera ray
	mat3 cam = camera();
	vec4 frand = hash(uvec4(frame)*uvec4(1,10,100,1000))*I2F;
	vec3 rp = vec3(0,0,cam[2].z),
		rd = normalize(vec3(uv.xy+(frand.xy*2.-1.)/screenY,cam[2].x));
	if (cam[1].x != 0.) {
		mat2 m = r2d(cam[1].x);
		rp.xy *= m; rd.xy *= m;
	}
	if (cam[1].y != 0.) {
		mat2 m = r2d(cam[1].y);
		rp.yz *= m; rd.yz *= m;
	}
	if (cam[1].z != 0.) {
		mat2 m = r2d(cam[1].z);
		rp.xz *= m; rd.xz *= m;
	}
	rp += cam[0].xyz;
	
	//ray trace view
	float eps = VIEW_EPSILON, t = cam[2].y;
	#ifdef VIEW_STEP_DITHER
	float dither = (1.-VIEW_STEP_DITHER)+VIEW_STEP_DITHER*frand.w;
	#endif
	baseRand = frand;
	for (int i = 0; i < VIEW_STEPS; i++) {
		vec2 geo = geometry(rp+rd*t);
		if (geo.x < eps) {
			gl_FragColor = vec4(normal(rp+rd*(t-eps)),t);
			return;
		}
		#ifdef VIEW_STEP_DITHER
		t += geo.x*dither;
		#else
		t += geo.x;
		#endif

		#ifdef VIEW_EPSILON_MUL
		eps *= VIEW_EPSILON_MUL;
		#endif
	}
	
	//no hit
	gl_FragColor = vec4(0,0,0,-1);
}
