#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2022

in vec3 in_Position;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};

vec4 baseRand;
float nsampling;

#include "art/PathTracerScene.glsl"


uniform sampler2D Buffer0, Buffer1, Buffer2;

void vert() {
	#ifdef ANIMATION_TIME
	time = mod(float(frame)/float(framerate*ANIMATION_SAMPLES),ANIMATION_TIME);
	#endif
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	ivec2 iuv = ivec2(uv.zw);
	
	//sample view buffer
	vec4 viewHit = texelFetch(Buffer0,iuv,0), col;
	if (viewHit.w >= 0.) {
		//setup camera ray
		mat3 cam = camera();
		vec4 frand = hash(uvec4(frame)*uvec4(1,10,100,1000))*I2F;
		vec3 rp = vec3(0,0,cam[2].z),
			rd = normalize(vec3(uv.xy+(frand.xy*2.-1.)/screenY,cam[2].x));
		if (cam[1].x != 0.) {
			mat2 m = r2d(cam[1].x);
			rp.xy *= m; rd.xy *= m;
		}
		if (cam[1].y != 0.) {
			mat2 m = r2d(cam[1].y);
			rp.yz *= m; rd.yz *= m;
		}
		if (cam[1].z != 0.) {
			mat2 m = r2d(cam[1].z);
			rp.xz *= m; rd.xz *= m;
		}
		rp += cam[0].xyz;
	
		//step forward to hit and sample material
		rp += rd*viewHit.w;
		
		//surface blur denoising
		col = texelFetch(Buffer1,iuv,0);
		float rayType = col.w;
		col.w = 1.;
		for (int y = -DENOISE_RANGE; y < DENOISE_RANGE+1; y++) {
			for (int x = -DENOISE_RANGE; x < DENOISE_RANGE+1; x++) {
				vec2 ouv = uv.zw+vec2(x,y);
				if ((x == 0 && y == 0) || min(ouv.x,ouv.y) < 0. || max(ouv.x-screenX,ouv.y-screenY) > 0.) continue;
				ivec2 iouv = ivec2(ouv);
				vec4 light = texelFetch(Buffer1,iouv,0);
				if (light.w != rayType) continue;
				light.w = 1.;
				vec4 hit = texelFetch(Buffer0,iouv,0);
				#ifdef DENOISE_NORMAL_THRESHOLD
				col += light*(step(max(viewHit.w,hit.w)/min(viewHit.w,hit.w),DENOISE_DEPTH_THRESHOLD)*max(0.,dot(viewHit.xyz,hit.xyz)-DENOISE_NORMAL_THRESHOLD)*(1./(1.-DENOISE_NORMAL_THRESHOLD)));
				#else
				col += light*(step(max(viewHit.w,hit.w)/min(viewHit.w,hit.w),DENOISE_DEPTH_THRESHOLD)*max(0.,dot(viewHit.xyz,hit.xyz)));
				#endif
			}
		}
		col /= col.w;
		
		//apply material color and emission
		baseRand = frand;
		mat3 mat = material(rp,viewHit.xyz,geometry(rp).y);
		col.xyz = col.xyz*mat[0]+mat[0]*mat[1].y;
		
	} else {
		//background
		col = texelFetch(Buffer1,iuv,0);
		col.w = 1.;
	}
	
	//accumulate samples
	#ifdef ANIMATION_TIME
	if (frame%ANIMATION_SAMPLES == 0) gl_FragColor = col;
	else gl_FragColor = texelFetch(Buffer2,iuv,0)+col;
	#else
	gl_FragColor = texelFetch(Buffer2,iuv,0)+col;
	#endif
}
